@@ -14,11 +14,13 @@ import requests
 import lxml.etree
 import xml2rfc.log
 import xml2rfc.utils
+from xml2rfc.writers import base
 
 try:
     from urllib.parse import urlparse, urljoin
 except ImportError:
     from urlparse import urlparse, urljoin
+
 try:
     import debug
     assert debug
@@ -44,24 +46,26 @@ class XmlRfcError(Exception):
 class CachingResolver(lxml.etree.Resolver):
     """ Custom ENTITY request handler that uses a local cache """
     def __init__(self, cache_path=None, library_dirs=None, source=None,
-                 templates_path='templates', verbose=False, quiet=False,
-                 no_network=False, network_locs= [
+                 templates_path='templates', verbose=None, quiet=None,
+                 no_network=None, network_locs= [
                      'https://xml2rfc.ietf.org/public/rfc/',
                      'https://xml2rfc.tools.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
                  ],
-                 rfc_number=None):
-        self.verbose = verbose
-        self.quiet = quiet
-        self.source = source
+                 rfc_number=None, options=base.default_options):
+        self.quiet = quiet if quiet != None else options.quiet
+        self.verbose = verbose if verbose != None else options.verbose
+        self.no_network = no_network if no_network != None else options.no_network
+        self.cache_path = cache_path or options.cache
+        self.source = source 
         self.library_dirs = library_dirs
         self.templates_path = templates_path
-        self.no_network=no_network
         self.network_locs = network_locs
         self.include = False
         self.rfc_number = rfc_number
         self.cache_refresh_secs = (60*60*24*14) # 14 days
+        self.options = options
 
         # Get directory of source
         if self.source:
@@ -75,9 +79,9 @@ class CachingResolver(lxml.etree.Resolver):
         # Determine cache directories to read/write to
         self.read_caches = [ os.path.expanduser(path) for path in xml2rfc.CACHES ]
         self.write_cache = None
-        if cache_path:
+        if self.cache_path:
             # Explicit directory given, set as first directory in read_caches
-            self.read_caches.insert(0, cache_path)
+            self.read_caches.insert(0, self.cache_path)
         # Try to find a valid directory to write to by stepping through
         # Read caches one by one
         for dir in self.read_caches:
@@ -114,6 +118,7 @@ class CachingResolver(lxml.etree.Resolver):
                 shutil.rmtree(path)
                 xml2rfc.log.write('Deleted cache directory at', path)
 
+                
     def resolve(self, request, public_id, context):
         """ Called internally by lxml """
         if not request:
@@ -134,9 +139,12 @@ class CachingResolver(lxml.etree.Resolver):
                 request = request[8:]
             try:
                 request = os.path.relpath(request, self.source_dir)
-            except ValueError:
-                pass
-        path = self.getReferenceRequest(request)
+            except ValueError as e:
+                xml2rfc.log.error(str(e))                
+        try:
+            path = self.getReferenceRequest(request)
+        except Exception as e:
+            xml2rfc.log.error(str(e))
         return self.resolve_filename(path, context)
     
     def getReferenceRequest(self, request, include=False, line_no=0):
@@ -147,11 +155,12 @@ class CachingResolver(lxml.etree.Resolver):
             If REQUEST ends with '.dtd' or '.ent' then
               If REQUEST is an absolute path (local or network) then
                 Return REQUEST
+              Else
+                Try TEMPLATE_DIR + REQUEST, otherwise
+                Return SOURCE_DIR + REQUEST
             Else
-              Try TEMPLATE_DIR + REQUEST, otherwise
-              Return SOURCE_DIR + REQUEST
-            Else
-              If REQUEST doesn't end with '.xml' then append '.xml'
+              If REQUEST doesn't end with '.xml' then
+                append '.xml'
               If REQUEST is an absolute path (local or network) then
                 Return REQUEST
               Else
@@ -204,10 +213,12 @@ class CachingResolver(lxml.etree.Resolver):
                     result = os.path.join(self.source_dir, basename)
                     attempts.append(result)
         else:
-            if request.endswith('.xml'):
+            if self.options and self.options.vocabulary == 'v3':
                 paths = [ request ]
-            else:
+            elif not request.endswith('.xml'):
                 paths = [ request, request + '.xml' ]
+            else:
+                paths = [ request ]                
             if os.path.isabs(paths[0]):
                 # Absolute path, return as-is
                 for path in paths:
@@ -235,8 +246,11 @@ class CachingResolver(lxml.etree.Resolver):
                         result = self.cache(path)
                         if result:
                             break
-                if not result and self.no_network:
-                    xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
+                if not result:
+                    if self.options.vocabulary == 'v3' and not request.endswith('.xml'):
+                        xml2rfc.log.warn("The v3 formatters require full explicit URLs of external resources.  Did you forget to add '.xml' (or some other extension)?")
+                    if self.no_network:
+                        xml2rfc.log.warn("Document not found in cache, and --no-network specified -- couldn't resolve %s" % request)
                 tried_cache = True
             else:
                 if os.path.dirname(paths[0]):
@@ -349,12 +363,19 @@ class CachingResolver(lxml.etree.Resolver):
         r = session.get(url)
         if r.status_code == 200:
             if self.write_cache:
-                write_path = os.path.join(self.write_cache, 
-                                          xml2rfc.CACHE_PREFIX, basename)
-                with codecs.open(write_path, 'w', encoding='utf-8') as cache_file:
-                    cache_file.write(r.text)
-                xml2rfc.log.note('Added file to cache: ', write_path)
-                return write_path
+                try:
+                    xml = lxml.etree.fromstring(r.text.encode('utf8'))
+                    xml.set('{%s}base'%xml2rfc.utils.namespaces['xml'], url)
+                    text = lxml.etree.tostring(xml, encoding='utf-8')
+                    write_path = os.path.join(self.write_cache, 
+                                              xml2rfc.CACHE_PREFIX, basename)
+                    with codecs.open(write_path, 'w', encoding='utf-8') as cache_file:
+                        cache_file.write(text.decode('utf-8'))
+                    xml2rfc.log.note('Added file to cache: ', write_path)
+                    return write_path
+                except Exception as e:
+                    xml2rfc.log.error(str(e))
+                    return url
             else:
                 return url
         else:
@@ -365,6 +386,12 @@ class CachingResolver(lxml.etree.Resolver):
 
 class AnnotatedElement(lxml.etree.ElementBase):
     pis = None
+    def get(self, key, default=None):
+        value = super(AnnotatedElement, self).get(key, default)
+        if value == default:
+            return value
+        else:
+            return six.text_type(value)
 
 class XmlRfcParser:
 
@@ -374,22 +401,31 @@ class XmlRfcParser:
 
 
     """ XML parser container with callbacks to construct an RFC tree """
-    def __init__(self, source, verbose=False, quiet=False, options=None,
+    def __init__(self, source, verbose=None, quiet=None, options=base.default_options,
                  cache_path=None, templates_path=None, library_dirs=None,
-                 no_network=False, network_locs=[
+                 no_network=None, network_locs=[
                      'https://xml2rfc.ietf.org/public/rfc/',
                      'https://xml2rfc.tools.ietf.org/public/rfc/',
                      'http://xml2rfc.ietf.org/public/rfc/',
                      'http://xml2rfc.tools.ietf.org/public/rfc/',
                  ]
                  ):
-        self.verbose = verbose
-        self.quiet = quiet
+        self.options = options
+        self.quiet = quiet if quiet != None else options.quiet
+        self.verbose = verbose if verbose != None else options.verbose
+        self.no_network = no_network if no_network != None else options.no_network
         self.source = source
-        self.cache_path = cache_path
-        self.no_network = no_network
+        self.cache_path = cache_path or options.cache
         self.network_locs = network_locs
 
+        if self.source:
+            if six.PY2:
+                with open(self.source, "rU") as f:
+                    self.text = f.read()
+            else:
+                with open(self.source, "rb", newline=None) as f:
+                    self.text = f.read()
+
         # Initialize templates directory
         self.templates_path = templates_path or \
                               os.path.join(os.path.dirname(xml2rfc.__file__),
@@ -427,6 +463,7 @@ class XmlRfcParser:
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
+                                        options=options,
                                     )
 
     def delete_cache(self, path=None):
@@ -437,13 +474,6 @@ class XmlRfcParser:
         if not (self.quiet or quiet):
             xml2rfc.log.write('Parsing file', self.source)
 
-        if six.PY2:
-            with open(self.source, "rU") as f:
-                self.text = f.read()
-        else:
-            with open(self.source, "rb", newline=None) as f:
-                self.text = f.read()
-
         # Get an iterating parser object
         file = six.BytesIO(self.text)
         context = lxml.etree.iterparse(file,
@@ -468,16 +498,19 @@ class XmlRfcParser:
                                         network_locs=self.network_locs,
                                         verbose=self.verbose,
                                         quiet=self.quiet,
+                                        options=self.options,
                                      )
         context.resolvers.add(caching_resolver)
 
         # Get hold of the rfc number (if any) in the rfc element, so we can
         # later resolve the "&rfc.number;" entity.
         self.rfc_number = None
+        self.format_version = None
         try:
             for action, element in context:
                 if element.tag == "rfc":
                     self.rfc_number = element.attrib.get("number", None)
+                    self.format_version = element.attrib.get("version", None)
                     break
         except lxml.etree.XMLSyntaxError as e:
             xml2rfc.log.warn("Parsing Error: %s" % e)
@@ -485,6 +518,9 @@ class XmlRfcParser:
             if e.message=="I/O operation on closed file":
                 pass
 
+        if self.format_version == "3":
+            self.default_dtd_path = None
+
         # now get a regular parser, and parse again, this time resolving entities
         parser = lxml.etree.XMLParser(dtd_validation=False,
                                       load_dtd=True,
@@ -506,6 +542,7 @@ class XmlRfcParser:
                                         verbose=self.verbose,
                                         quiet=self.quiet,
                                         rfc_number = self.rfc_number,
+                                        options=self.options,
                                     )
 
         # Add our custom resolver
