@@ -1,20 +1,20 @@
-# --------------------------------------------------
-# Copyright The IETF Trust 2011, All Rights Reserved
-# --------------------------------------------------
+# Copyright The IETF Trust 2018, All Rights Reserved
+# -*- coding: utf-8 -*-
+from __future__ import unicode_literals, print_function, division
 
 # Internal utitlity functions.  Not meant for public usage.
 
 import base64
-import calendar
 import re
 import six
 import textwrap
 
+from lxml.etree import _Comment
+
 if six.PY2:
     from urllib import quote
 else:
     from urllib.request import quote
-    
 
 try:
     import debug
@@ -24,7 +24,11 @@ except ImportError:
 
 import xml2rfc.log
 
-class MyTextWrapper(textwrap.TextWrapper):
+
+# ----------------------------------------------------------------------
+# Text wrapping
+
+class TextWrapper(textwrap.TextWrapper):
     """ Subclass that overrides a few things in the standard implementation """
     def __init__(self, **kwargs):
         textwrap.TextWrapper.__init__(self, **kwargs)
@@ -103,14 +107,20 @@ class MyTextWrapper(textwrap.TextWrapper):
             text = re.sub(re.escape(key), val, text)
         return text
 
-    def wrap(self, text, initial_indent='', subsequent_indent='',
+    def wrap(self, text, initial='', subsequent_indent=None, width=None,
         fix_doublespace=True, fix_sentence_endings=True, drop_whitespace=True):
         """ Mirrored implementation of wrap which replaces characters properly
-            also lets you easily specify indentation on the fly
+            and also lets you easily specify indentation on the fly
         """
         # Set indentation
-        self.initial_indent = initial_indent
-        self.subsequent_indent = subsequent_indent
+        _width = self.width
+        if width != None:
+            self.width = width
+        self.initial_indent = initial
+        if subsequent_indent == None:
+            self.subsequent_indent = initial
+        else:
+            self.subsequent_indent = subsequent_indent
         self.fix_sentence_endings = fix_sentence_endings
         self.drop_whitespace = drop_whitespace
 
@@ -119,7 +129,7 @@ class MyTextWrapper(textwrap.TextWrapper):
 
         # Maybe remove double (and more) spaces, except when they might be between sentences
         if fix_doublespace:
-            text = re.sub("([^].!?])  +", r"\1 ", text)
+            text = re.sub("([^.!?])  +", r"\1 ", text)
             text = re.sub("([.!?])   +", r"\1  ", text)
 
         # prevent breaking "Section N.N" and "Appendix X.X"
@@ -128,7 +138,7 @@ class MyTextWrapper(textwrap.TextWrapper):
         # Replace some characters after splitting has occured
         parts = self._split(text)
         chunks = []
-        max_word_len = self.width - len(subsequent_indent)
+        max_word_len = self.width - len(self.subsequent_indent)
         for chunk in parts:
             chunk2 = self.replace(chunk)
             if len(chunk2) > max_word_len:
@@ -146,10 +156,12 @@ class MyTextWrapper(textwrap.TextWrapper):
         # Original implementation
         if self.fix_sentence_endings:
             self._fix_sentence_endings(chunks)
-        return self._wrap_chunks(chunks)
+        wrapped = self._wrap_chunks(chunks)
+        self.width = _width
+        return wrapped
 
     def fill(self, *args, **kwargs):
-        return "\n".join(self.wrap(*args, **kwargs))
+        return "\n".join(self.wrap(*args, **kwargs)).replace(u'\u00A0', ' ')
 
 
 def justify_inline(left_str, center_str, right_str, width=72):
@@ -181,24 +193,9 @@ def justify_inline(left_str, center_str, right_str, width=72):
             output[i] = right[i]
     return ''.join(output)
 
-def formatXmlWhitespace(tree):
-    """ Traverses an lxml.etree ElementTreeand properly formats whitespace
-    
-        We replace newlines with single spaces, unless it ends with a
-        period then we replace the newline with two spaces.
-    """
-    for element in tree.iter():
-        # Preserve formatting on artwork
-        if element.tag != 'artwork':
-            if element.text is not None:
-                element.text = re.sub('\s*\n\s*', ' ', \
-                               re.sub('\.\s*\n\s*', '.  ', \
-                               element.text.lstrip()))
+# ----------------------------------------------------------------------
+# Other text operations
 
-            if element.tail is not None:
-                element.tail = re.sub('\s*\n\s*', ' ', \
-                               re.sub('\.\s*\n\s*', '.  ', \
-                               element.tail))
 
 def ascii_split(text):
     """ We have unicode strings, but we want to split only on the ASCII
@@ -210,61 +207,6 @@ def ascii_split(text):
     return re.split("[ \t\n\r\f\v]+", text)
 
 
-# ----------------------------------------------------------------------
-# Base conversions.
-# From http://tech.vaultize.com/2011/08/python-patterns-number-to-base-x-and-the-other-way/
-
-DEFAULT_DIGITS = '0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ'
- 
-def num_to_baseX(num, digits=DEFAULT_DIGITS):
-   if num < 0: return '-' + num_to_baseX(-num)
-   if num == 0: return digits[0]
-   X = len(digits)
-   s = ''
-   while num > 0:
-        s = digits[num % X] + s
-        num //= X
- 
-   return s
- 
-def baseX_to_num(s, digits=DEFAULT_DIGITS):
-   if s[0] == '-': return -1 * baseX_to_num(s[1:])
-   ctopos = dict([(c, pos) for pos, c in enumerate(digits)])
-   X = len(digits)
-   num = 0
-   for c in s: num = num * X + ctopos[c]
-   return num
-
-# ----------------------------------------------------------------------
-# Use the generic base conversion to create list letters
-
-def int2letter(num):
-    return num_to_baseX(num, "abcdefghijklmnopqrstuvwxyz")
-
-def int2roman(number):
-    numerals = { 
-        1 : "i", 
-        4 : "iv", 
-        5 : "v", 
-        9 : "ix", 
-        10 : "x", 
-        40 : "xl", 
-        50 : "l", 
-        90 : "xc", 
-        100 : "c", 
-        400 : "cd", 
-        500 : "d", 
-        900 : "cm", 
-        1000 : "m" 
-    }
-    result = ""
-    for value, numeral in sorted(numerals.items(), reverse=True):
-        while number >= value:
-            result += numeral
-            number -= value
-    return result
-
-
 def urlkeep(text):
     """ Insert word join XML entities on forward slashes and hyphens
         in a URL so that it stays on one line
@@ -283,6 +225,29 @@ def urlkeep(text):
         return re.sub('(?<=https:)\S*', replacer, text)
     return text
 
+# ----------------------------------------------------------------------
+# Tree operations
+
+def formatXmlWhitespace(tree):
+    """ Traverses an lxml.etree ElementTree and properly formats whitespace
+    
+        We replace newlines with single spaces, unless it ends with a
+        period then we replace the newline with two spaces.
+    """
+    for element in tree.iter():
+        # Preserve formatting on artwork
+        if element.tag != 'artwork':
+            if element.text is not None:
+                element.text = re.sub('\s*\n\s*', ' ', \
+                               re.sub('\.\s*\n\s*', '.  ', \
+                               element.text.lstrip()))
+
+            if element.tail is not None:
+                element.tail = re.sub('\s*\n\s*', ' ', \
+                               re.sub('\.\s*\n\s*', '.  ', \
+                               element.tail))
+
+
 def safeReplaceUnicode(tree):
     """ Traverses an lxml.etree ElementTree and replaces unicode characters 
         with the proper equivalents specified in rfc2629-xhtml.ent.
@@ -309,6 +274,52 @@ def safeReplaceUnicode(tree):
                 _replace_unicode_characters(element.attrib[key])
 
 
+def safeTagSlashedWords(tree):
+    """ Traverses an lxml.etree ElementTree and replace words seperated
+        by slashes if they are on the list
+    """
+    slashList = {}
+    for element in _slash_replacements:
+        slashList[element] = re.sub(u'/', u'\u200B/\u200B', element)
+        
+    for element in tree.iter():
+        if element.text:
+            element.text = _replace_slashed_words(element.text, slashList)
+        if element.tail:
+            element.tail = _replace_slashed_words(element.tail, slashList)
+        # I do not know that this makes any sense
+        # for key in element.attrib.keys():
+        #    element.attrib[key] = \
+        #        _replace_slashed_words(element.attrib[key], slashList)
+
+
+def find_duplicate_ids(schema, tree):
+    # get attributes specified with data type "ID"
+    id_data = schema.xpath("/x:grammar/x:define/x:element//x:attribute/x:data[@type='ID']", namespaces=namespaces)
+    attr = set([ i.getparent().get('name') for i in id_data ])
+    # Check them one by one
+    return find_duplicate_attr_values(attr, tree)
+
+def find_duplicate_html_ids(tree):
+    return find_duplicate_attr_values(['id',], tree)
+
+def find_duplicate_attr_values(attr, tree):
+    dups = []
+    # Check them one by one
+    for a in attr:
+        seen = set()
+        for e in tree.xpath('.//*[@%s]' % a):
+            id = e.get(a)
+            if id != None and id in seen:
+                dups.append((a, id, e))
+            else:
+                seen.add(id)
+    return dups
+
+# ----------------------------------------------------------------------
+# Unicode operations
+
+
 def _replace_unicode_characters(str):
     """ replace those Unicode characters that we do not use internally
         &wj; &zwsp; &nbsp; &nbhy;
@@ -329,6 +340,7 @@ def _replace_unicode_characters(str):
 # Auto-generated from comments in rfc2629-xhtml.ent
 _unicode_replacements = {
     u'\x09': ' ',
+    u'\xa0': ' ',
     u'\xa1': '!',
     u'\xa2': '[cents]',
     u'\xa3': 'GBP',
@@ -512,24 +524,6 @@ def parse_pi(pi, pis):
         return tmp_dict
     return {}
 
-def safeTagSlashedWords(tree):
-    """ Traverses an lxml.etree ElementTree and replace words seperated
-        by slashes if they are on the list
-    """
-    slashList = {}
-    for element in _slash_replacements:
-        slashList[element] = re.sub(u'/', u'\u200B/\u200B', element)
-        
-    for element in tree.iter():
-        if element.text:
-            element.text = _replace_slashed_words(element.text, slashList)
-        if element.tail:
-            element.tail = _replace_slashed_words(element.tail, slashList)
-        # I do not know that this makes any sense
-        # for key in element.attrib.keys():
-        #    element.attrib[key] = \
-        #        _replace_slashed_words(element.attrib[key], slashList)
-
 
 def _replace_slashed_words(str, slashList):
     """ replace slashed separated words the list with
@@ -561,12 +555,37 @@ def build_dataurl(mime, data, base64enc=False):
         url = "data:%s,%s" % (mime, data)
     return url
 
-def normalize_month(month):
-    if len(month) < 3:
-        xml2rfc.log.error("Expected a month name with at least 3 letters, found '%s'" % (month, ))
-    for i, m in enumerate(calendar.month_name):
-        if m and m.lower().startswith(month.lower()):
-            month = '%02d' % (i)
-    assert month.isdigit()
-    return month
+namespaces={
+    'x':    'http://relaxng.org/ns/structure/1.0',
+    'a':    'http://relaxng.org/ns/compatibility/annotations/1.0',
+    'xml':  'http://www.w3.org/XML/1998/namespace',
+    's':    'http://www.w3.org/2000/svg',
+}
 
+
+# ----------------------------------------------------------------------
+# Element operations
+
+
+def isempty(e):
+    "Return True if e has no children and no text content"
+    return not ((len(e) and any([ not isinstance(c, _Comment) for c in e.iterchildren() ]))
+                or (e.text and e.text.strip()) or (e.tail and e.tail.strip()))
+
+def isblock(e):
+    "Return True if e is a block level element"
+    return e.tag in [ 'artwork', 'dl', 'figure', 'ol', 'sourcecode', 't', 'ul', 'blockquote', ]
+
+def iscomment(e):
+    "Return True if e is a comment"
+    return isinstance(e, _Comment)
+
+def hastext(e):
+    "Return a list of text-level immediate children"
+    head = [ e.text ] if e.text and e.text.strip() else []
+    items = head + [ c for c in e.iterchildren() if not (isblock(c) or iscomment(c))] + [ c.tail for c in e.iterchildren() if c.tail and c.tail.strip() ]
+    return items
+
+def is_htmlblock(h):
+    return h.tag in set([ 'address', 'article', 'aside', 'blockquote', 'div', 'dl', 'figure',
+        'h1', 'h2', 'h3', 'h4', 'h5', 'h6', 'nav', 'ol', 'p', 'pre', 'section', 'table', 'ul', ])
