@@ -3,12 +3,18 @@ from xml2rfc.writers.base import RfcWriterError
 from xml2rfc.writers.base import BaseRfcWriter
 from xml2rfc.writers.raw_txt import RawTextRfcWriter
 from xml2rfc.writers.paginated_txt import PaginatedTextRfcWriter
-from xml2rfc.writers.html import HtmlRfcWriter
+from xml2rfc.writers.legacy_html import HtmlRfcWriter
 from xml2rfc.writers.nroff import NroffRfcWriter
 from xml2rfc.writers.expanded_xml import ExpandedXmlWriter
 from xml2rfc.writers.v2v3 import V2v3XmlWriter
 from xml2rfc.writers.preptool import PrepToolWriter
+from xml2rfc.writers.text import TextWriter
+from xml2rfc.writers.html import HtmlWriter
+from xml2rfc.writers.expand import ExpandV3XmlWriter
 
+# This defines what 'from xml2rfc.writers import *' actually imports:
 __all__ = ['BaseRfcWriter', 'RawTextRfcWriter', 'PaginatedTextRfcWriter',
            'HtmlRfcWriter', 'NroffRfcWriter', 'ExpandedXmlWriter',
-           'RfcWriterError', 'V2v3XmlWriter', 'PrepToolWriter', ]
+           'RfcWriterError', 'V2v3XmlWriter', 'PrepToolWriter', 'TextWriter',
+           'HtmlWriter', 'ExpandV3XmlWriter',
+       ]
