@@ -7,18 +7,57 @@ import codecs
 import datetime
 import textwrap
 import lxml
+import os
 import re
+import sys
 import xml2rfc.log
+import xml2rfc.util
 import xml2rfc.utils
 from optparse import Values
 
 try:
     import debug
-    assert debug
+    debug.debug = True
 except ImportError:
     pass
 
-default_options = Values(defaults=dict(quiet=False, verbose=False, utf8=False, debug=False, liberal=False))
+from xml2rfc import strings
+from xml2rfc.util.date import extract_date, format_date, get_expiry_date
+from xml2rfc.util.name import short_author_ascii_name_parts, full_author_name_expansion
+from xml2rfc.util.unicode import punctuation, unicode_replacements, unicode_content_tags, downcode
+from xml2rfc.utils import namespaces
+
+default_options = Values(defaults={
+        'accept_prepped': None,
+        'add_xinclude': None,
+        'basename': None,
+        'cache': None,
+        'css': None,
+        'datestring': None,
+        'debug': False,
+        'doi_base_url': 'https://doi.org/',
+        'dtd': None,
+        'external_css': False,
+        'filename': None,
+        'id_base_url': 'https://www.ietf.org/archive/id/',
+        'indent': 2,
+        'legacy': False,
+        'legacy_date_format': True,
+        'legacy_list_symbols': False,
+        'list_symbols': ('*', '-', 'o', '+'),
+        'no_dtd': None,
+        'no_network': False,
+        'omit_headers': None,
+        'output_filename': None,
+        'output_path': None,
+        'quiet': False,
+        'remove_pis': False,
+        'rfc_base_url': 'https://www.rfc-editor.org/info/',
+        'strict': False,
+        'utf8': False,
+        'verbose': False,
+        'vocabulary': 'v2',
+    })
 
 class _RfcItem:
     """ A unique ID object for an anchored RFC element.
@@ -428,6 +467,8 @@ class BaseRfcWriter:
             item = _RfcItem(autoName, autoAnchor, counter=counter, title=title,
                            anchor=anchor, toc=toc, level=level, appendix=appendix)
         else:
+            if not title:
+                raise RfcWriterError("No title available when trying to insert index item for anchor %s" % anchor)
             autoAnchor = 'rfc.' + re.sub('[^A-Za-z0-9]+', '_', title).lower()
             item = _RfcItem(title, autoAnchor, title=title)
         self._index.append(item)
@@ -483,6 +524,13 @@ class BaseRfcWriter:
         self._index.append(item)
         return item
 
+    def _indexAuthor(self, counter, anchor, name):
+        autoName = name
+        autoAnchor = 'rfc.author.%s' % counter
+        item = _RfcItem(autoName, autoAnchor, counter=counter, anchor=anchor, toc=False)
+        self._index.append(item)
+        return item
+
     def get_initials(self, author):
         """author is an rfc2629 author element.  Return the author initials,
         fixed up according to current flavour and policy."""
@@ -559,7 +607,7 @@ class BaseRfcWriter:
             date.set('month', str(month))
         else:
             if not month.isdigit():
-                month = xml2rfc.utils.normalize_month(month)
+                month = xml2rfc.util.date.normalize_month(month)
             month = int(month)
         #
         day = date.attrib.get('day')
@@ -616,13 +664,13 @@ class BaseRfcWriter:
         if '%d' in text:
             text = text.replace(r'%d', str(count)).ljust(decimal_width+extra_width)
         elif '%c' in text:
-            text = text.replace(r'%c', xml2rfc.utils.int2letter(count-1)).ljust(letter_width+extra_width)
+            text = text.replace(r'%c', xml2rfc.util.num.int2letter(count)).ljust(letter_width+extra_width)
         elif '%C' in text:
-            text = text.replace(r'%C', xml2rfc.utils.int2letter(count-1).upper()).ljust(letter_width+extra_width)
+            text = text.replace(r'%C', xml2rfc.util.num.int2letter(count).upper()).ljust(letter_width+extra_width)
         elif '%i' in text:
-            text = text.replace(r'%i', xml2rfc.utils.int2roman(count)).ljust(roman_width+extra_width)
+            text = text.replace(r'%i', xml2rfc.util.num.int2roman(count)).ljust(roman_width+extra_width)
         elif '%I' in text:
-            text = text.replace(r'%I', xml2rfc.utils.int2roman(count).upper()).ljust(roman_width+extra_width)
+            text = text.replace(r'%I', xml2rfc.util.num.int2roman(count).upper()).ljust(roman_width+extra_width)
         elif '%o' in text:
             text = text.replace(r'%o', oct(count).replace("0","",1)).replace("o","",1).ljust(octal_width+extra_width)
         elif '%x' in text:
@@ -739,7 +787,9 @@ class BaseRfcWriter:
         # Keep track of previous organization and remove if redundant.
         last_org = None
         last_pos = None
-        for author in self.r.findall('front/author'):
+        authors = self.r.findall('front/author')
+        authors = [ a for a in authors if a.get('role') != 'contributor' ]
+        for author in authors:
             role = author.attrib.get('role', '')
             if role == 'editor':
                 role = ', Ed.'
@@ -1005,8 +1055,8 @@ class BaseRfcWriter:
             category = self.r.attrib.get('category', 'none')
             stream = self.r.attrib.get('submissionType', 
                                        self.defaults['submissionType'])
-            # Consensus is assumed 'yes' for Standards Track documents
-            consensus = category == 'std' and 'yes' or \
+            # Consensus is assumed 'yes' for Standards Track documents and BCPs
+            consensus = category in ['std', 'bcp'] and 'yes' or \
                         self.r.attrib.get('consensus', self.defaults['consensus'])
             workgroup = ''
             wg_element = self.r.find('front/workgroup')
@@ -1129,6 +1179,14 @@ class BaseRfcWriter:
             for t in abstract.findall('t'):
                 self.write_t_rec(t)
 
+        # Authors
+        count = 0
+        authors = self.r.findall('front/author')
+        for author in authors:
+            count += 1
+            autoName = full_author_name_expansion(author)
+            self._indexAuthor(count, author.get('anchor'), autoName)
+
         # Optional notes
         for note in self.r.findall('front/note'):
             self.write_heading(note.attrib.get('title', 'Note'))
@@ -1190,6 +1248,7 @@ class BaseRfcWriter:
 
         if self.pis['authorship'] == 'yes':
             authors = self.r.findall('front/author')
+            authors = [ a for a in authors if a.get('role') != 'contributor' ]
             autoAnchor = 'rfc.authors'
             if len(authors) > 1:
                 title = "Authors' Addresses"
@@ -1315,6 +1374,7 @@ class BaseRfcWriter:
         # Authors addresses section
         if self.pis['authorship'] == 'yes':
             authors = self.r.findall('front/author')
+            authors = [ a for a in authors if a.get('role') != 'contributor' ]
             autoAnchor = 'rfc.authors'
             if len(authors) > 1:
                 title = "Authors' Addresses"
@@ -1472,3 +1532,284 @@ class BaseRfcWriter:
     def write_to_file(self, file):
         """ Writes the finished buffer to a file """
         raise NotImplementedError('write_to_file() needs to be overridden')
+
+
+        
+deprecated_element_tags = set([
+    'c',
+    'facsimile',
+    'format',
+    'list',
+    'postamble',
+    'preamble',
+    'spanx',
+    'texttable',
+    'ttcol',
+    'vspace',
+])
+
+
+class BaseV3Writer(object):
+
+    def __init__(self, xmlrfc, quiet=None, options=default_options, date=datetime.date.today()):
+        self.xmlrfc = xmlrfc
+        self.tree = xmlrfc.tree
+        self.root = self.tree.getroot()
+        self.options = options
+        self.date = date
+        self.index_items = []
+        self.v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
+        self.schema = lxml.etree.ElementTree(file=self.v3_rng_file)
+        self.text_tags   = self.get_text_tags() - deprecated_element_tags
+        self.inline_tags = self.get_inline_tags() - deprecated_element_tags
+        self.refname_mapping = dict( (e.get('anchor'), e.get('anchor')) for e in self.root.xpath('.//reference') )
+        self.refname_mapping.update(dict( (e.get('target'), e.get('to')) for e in self.root.xpath('.//displayreference') ))
+        #
+        self.errors = []
+        if options.debug:
+            found_handlers = []
+            missing_handlers = []
+            for tag in self.element_tags:
+                func_name = "render_%s" % (tag.lower(),)
+                if getattr(self, func_name, False):
+                    found_handlers.append(func_name)
+                else:
+                    missing_handlers.append(func_name)
+            debug.pprint('found_handlers')
+            debug.show('len(found_handlers)')
+            debug.pprint('missing_handlers')
+            debug.show('len(missing_handlers)')
+
+    def msg(self, e, label, text):
+        lnum = getattr(e, 'sourceline', None)
+        file = getattr(e, 'base', None)
+        if lnum:
+            msg = "%s(%s): %s %s" % (file or self.xmlrfc.source, lnum, label, text, )
+        else:
+            msg = "(No source line available): %s %s" % (label, text, )
+        xml2rfc.log.write(msg)
+        return msg
+
+    def get_relevant_pis(self, e):
+        pis = []
+        # directly inside element
+        for c in e.getchildren():
+            if c.tag == lxml.etree.PI and c.target == xml2rfc.V3_PI_TARGET:
+                pis.append(c)
+        # siblings before element
+        for s in e.itersiblings(preceding=True):
+            if s.tag == lxml.etree.PI and s.target == xml2rfc.V3_PI_TARGET:
+                pis.append(s)
+        # ancestor's earlier siblings
+        for a in e.iterancestors():
+            for s in a.itersiblings(preceding=True):
+                if s.tag == lxml.etree.PI and s.target == xml2rfc.V3_PI_TARGET:
+                    pis.append(s)
+        # before root elements
+        p = self.root.getprevious()
+        while p != None:
+            if p.tag == lxml.etree.PI and p.target == xml2rfc.V3_PI_TARGET:
+                pis.append(p)
+            p = p.getprevious()
+        return pis
+
+    def silenced(self, e, text):
+        text = text.strip()
+        pis = self.get_relevant_pis(e)
+        silenced = filter(None, [ pi.get('silence') for pi in pis ])
+        for s in silenced:
+            for label in ['Note: ', 'Warning: ']:
+                if s.startswith(label):
+                    s = s[len(label):]
+            if text.startswith(s):
+                return True
+            try:
+                if re.match(s, text):
+                    return True
+            except re.error:
+                pass
+        return False
+
+    def note(self, e, text, label='Note:'):
+        if not self.options.quiet:
+            if not self.silenced(e, text):
+                self.msg(e, label, text)
+
+    def warn(self, e, text):
+        if not self.silenced(e, text):
+            self.msg(e, 'Warning:', text)
+
+    def err(self, e, text, trace=False):
+        msg = self.msg(e, 'Error:', text)
+        if trace or self.options.debug:
+            raise RuntimeError(msg)
+        self.errors.append(msg)
+
+    def die(self, e, text, trace=False):
+        self.err(e, text, trace)
+        xml2rfc.log.write("Cannot continue, quitting now")
+        sys.exit(1)
+
+#     def get_valid_child_tags(self, tag):
+#         refs = self.schema.xpath("/x:grammar/x:define/x:element[@name='%s']//x:ref" % tag, namespaces=utils.namespaces)
+#         names = set([ r.get('name') for r in refs ])
+#         return names
+
+    def get_text_tags(self):
+        "Get tags that can have text content from the schema"
+        text_tags = set()
+        texts = self.schema.xpath("/x:grammar/x:define/x:element//x:text", namespaces=namespaces)
+        for t in texts:
+            element = list(t.iterancestors('{*}element'))[0]
+            name = element.get('name')
+            if not name in text_tags:
+                text_tags.add(name)
+        return text_tags
+
+    def get_inline_tags(self):
+        "Get tags that can occur within text from the schema"
+        inline_tags = set()
+        referenced = self.schema.xpath("/x:grammar/x:define/x:element//x:ref", namespaces=namespaces)
+        for ref in referenced:
+            name = ref.get('name')
+            if not name in inline_tags:
+                p = ref.getparent()
+                text = p.find('x:text', namespaces=namespaces)
+                if text != None:
+                    inline_tags.add(name)
+        return inline_tags
+
+    def write(self, filename):
+        raise NotImplementedError()
+
+    # methods for use in setting up page headers and footers
+
+    def footer_series(self):
+        if   self.root.find('.//seriesInfo[@name="Internet-Draft"]') != None:
+            return 'Internet-Draft'
+        else:
+            info = self.root.find('.//seriesInfo[@name="RFC"]')
+            if info != None:
+                return 'RFC %s' % info.get('value')
+        return '(unknown)'
+
+    def footer_title(self):
+        title = self.root.find('./front/title')
+        text = title.get('abbrev') or ''.join(title.itertext())
+        return text
+
+    def footer_date(self):
+        date = self.root.find('./front/date')
+        parts = extract_date(date, self.options.date)
+        text = format_date(parts[0], parts[1], None, legacy=True)
+        return text
+
+    def footer_authors(self):
+        authors = self.root.findall('front/author')
+        authors = [ a for a in authors if a.get('role') != 'contributor' ]
+        surnames = [ n for n in [ short_author_ascii_name_parts(a)[1] for a in authors ] if n ]
+        if len(surnames) == 1:
+            text = surnames[0]
+        elif len(surnames) == 2:
+            text = '%s & %s' % (surnames[0], surnames[1])
+        elif len(surnames) > 2:
+            text = '%s, et al.' % surnames[0]
+        return text
+
+    def footer_center(self):
+        # Either expiry date or category
+        if self.options.rfc:
+            cat = self.root.get('category')
+            text = strings.category_name[cat]
+        else:
+            date = get_expiry_date(self.tree, self.options.date)
+            parts = date.year, date.month, date.day
+            text = 'Expires %s' % format_date(*parts, legacy=self.options.legacy_date_format)
+        return text
+
+    def downcode_punctuation(self):
+        self.downcode(replacements=punctuation)
+
+    def downcode(self, replacements=unicode_replacements):
+        """
+
+        Traverses an lxml.etree and replaces unicode characters with the proper
+        equivalents specified in rfc2629-xhtml.ent, resulting in no non-ascii
+        characters except in elements that explicitly permit non-ascii content.
+
+        """
+        for e in self.tree.iter():
+            if e.text:
+                if not e.tag in unicode_content_tags:
+                    try:
+                        e.text.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.text = downcode(e.text, replacements=replacements)
+            if e.tail:
+                if not e.getparent().tag in unicode_content_tags:
+                    try:
+                        e.tail.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.tail = downcode(e.tail, replacements=replacements)
+            if e.tag != 'author':
+                for key in e.attrib.keys():
+                    try:
+                        e.get(key).encode('ascii')
+                    except UnicodeEncodeError:
+                        e.set(key, downcode(e.get(key), replacements=replacements))
+
+    def pretty_print_prep(self, e, p):
+        ind = self.options.indent
+        ## The actual printing is done in self.write()
+        def indent(e, i):
+            if e.tag in (lxml.etree.CDATA, ):
+                return
+            if e.tag in (lxml.etree.Comment, lxml.etree.PI, ):
+                if not e.tail:
+                    if e.getnext() != None:
+                        e.tail = '\n'+' '*i
+                    else:
+                        e.tail = '\n'+' '*(i-ind)
+                return
+            #
+            if e.tag not in self.text_tags:
+                if len(e) and (e.text is None or e.text.strip()==''):
+                    e.text = '\n'+' '*(i+ind)
+            elif e.tag in ['blockquote', 'li', 'dd', 'td', 'th' ]: # mixed content
+                pass
+                if len(e) and e[0] not in self.inline_tags and (e.text is None or e.text.strip()==''):
+                    e.text = '\n'+' '*(i+ind)
+            elif e.tag in ['artwork', 'sourcecode', ]:
+                pass
+            else:
+                # inline tag
+                if e.tail == '':
+                    e.tail = None
+                if len(e):
+                    z = e[-1]
+                    if z.tail and re.search(r'\n[ \t]+$', z.tail):
+                        z.tail = re.sub(r'\n[ \t]+$', '\n'+' '*(i), z.tail)
+                else:
+                    if e.text and re.search(r'\n[ \t]+$', e.text):
+                        e.text = re.sub(r'\n[ \t]+$', '\n'+' '*(i), e.text)
+                    
+            #
+            for c in e:
+                indent(c, i+ind)
+            #
+            if e.tag not in self.inline_tags:# and e.tag not in ['artwork', 'sourcecode', ]:
+                if e.tail is None or e.tail.strip()=='':
+                    if e.getnext() != None:
+                        e.tail = '\n'+' '*i
+                    else:
+                        e.tail = '\n'+' '*(i-ind)
+                else:
+                    self.warn(e, 'Unexpected tail: <%s>%s' % (e.tag, e.tail))
+            else:
+                if e.tail != None and e.tail.strip()=='' and '\n' in e.tail:
+                    if e.getnext() != None:
+                        e.tail = '\n'+' '*i
+                    else:
+                        e.tail = '\n'+' '*(i-ind)
+        indent(e, 0)
+        e.tail = None
