@@ -16,6 +16,7 @@ except ImportError:
 
 # Local lib
 from xml2rfc.writers.base import BaseRfcWriter, default_options
+from xml2rfc.util.unicode import expand_unicode_element
 import xml2rfc.utils
 
 
@@ -47,10 +48,10 @@ class RawTextRfcWriter(BaseRfcWriter):
         self.iref_marks = {}
 
         # Text lookups
-        self.inline_tags = ['xref', 'eref', 'iref', 'cref', 'spanx']
+        self.inline_tags = ['xref', 'eref', 'iref', 'cref', 'spanx', 'u', ]
         
         # Custom textwrapper object
-        self.wrapper = xml2rfc.utils.MyTextWrapper(width=self.width,
+        self.wrapper = xml2rfc.utils.TextWrapper(width=self.width,
                                                    fix_sentence_endings=True)
 
     def _length(self, text):
@@ -98,7 +99,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         subsequent = ' ' * (indent + sub_indent)
         if bullet:
             initial = ' ' * indent
-            bullet_parts = self.wrapper.wrap(bullet, initial_indent=initial, subsequent_indent=initial,
+            bullet_parts = self.wrapper.wrap(bullet, initial=initial,
                                              fix_doublespace=False, drop_whitespace=False)
             if len(bullet_parts) > 1:
                 buf.extend(bullet_parts[:-1])
@@ -115,11 +116,11 @@ class RawTextRfcWriter(BaseRfcWriter):
                 # Strip initial whitespace
                 string = string.lstrip()
             if wrap_urls:
-                par = self.wrapper.wrap(string, initial_indent=initial, subsequent_indent=subsequent,
+                par = self.wrapper.wrap(string, initial=initial, subsequent_indent=subsequent,
                                         fix_sentence_endings=fix_sentence_endings,
                                         fix_doublespace=fix_doublespace)
             else:
-                par = self.wrapper.wrap(xml2rfc.utils.urlkeep(string), initial_indent=initial,
+                par = self.wrapper.wrap(xml2rfc.utils.urlkeep(string), initial=initial,
                                         subsequent_indent=subsequent,
                                         fix_sentence_endings=fix_sentence_endings,
                                         fix_doublespace=fix_doublespace)
@@ -396,6 +397,14 @@ class RawTextRfcWriter(BaseRfcWriter):
         else:
             return target_text
 
+    def _expand_u(self, u):
+        try:
+            text = expand_unicode_element(u)
+        except (RuntimeError, ValueError) as exc:
+            text = ''
+            xml2rfc.log.error('%s'%exc)
+        return text
+
     def write_ref_element(self, key, text, sub_indent, source_line=None):
         """ Render a single reference element """
         # Use an empty first line if key is too long
@@ -447,26 +456,26 @@ class RawTextRfcWriter(BaseRfcWriter):
                     line.append('<' + element.attrib['target'].strip() + '>')
             elif element.tag == 'iref':
                 self._add_iref_to_index(element)
-            elif element.tag == 'cref' and \
-                element.pis['comments'] == 'yes':                
-                # Render if processing instruction is enabled
-                self.cref_counter += 1
-                anchor = element.attrib.get('anchor', None)
-                if anchor is None:
-                    anchor = 'CREF' + str(self.cref_counter)
-                    element.attrib['anchor'] = anchor
-                self._indexCref(self.cref_counter, anchor)
-                if element.pis['inline'] == 'yes':
-                    if anchor:
-                        anchor = anchor + ': '
-                    source = element.attrib.get('source', "")
-                    if source:
-                        source = " --" + source
-                    if element.text:
-                        line.append('[[' + anchor + element.text + source + ']]')
-                else:
-                    line.append('[' + anchor + ']')
-                    self.cref_list.append(element)
+            elif element.tag == 'cref':
+                if element.pis['comments'] == 'yes':                
+                    # Render if processing instruction is enabled
+                    self.cref_counter += 1
+                    anchor = element.attrib.get('anchor', None)
+                    if anchor is None:
+                        anchor = 'CREF' + str(self.cref_counter)
+                        element.attrib['anchor'] = anchor
+                    self._indexCref(self.cref_counter, anchor)
+                    if element.pis['inline'] == 'yes':
+                        if anchor:
+                            anchor = anchor + ': '
+                        source = element.attrib.get('source', "")
+                        if source:
+                            source = " --" + source
+                        if element.text:
+                            line.append('[[' + anchor + element.text + source + ']]')
+                    else:
+                        line.append('[' + anchor + ']')
+                        self.cref_list.append(element)
             elif element.tag == 'spanx':
                 style = element.attrib.get('style', 'emph')
                 edgechar = '_'  # default to emph because the spanx element exists
@@ -480,7 +489,11 @@ class RawTextRfcWriter(BaseRfcWriter):
                 if element.text:
                     text = element.text
                 line.append(edgechar + text + edgechar)
-            
+            elif element.tag == 'u':
+                line.append(self._expand_u(element))
+            else:
+                xml2rfc.log.error("Found unexpected inline element: <%s>" % element.tag)
+
             # Add tail text before next element
             if element.tail:
                 line.append(element.tail)
@@ -734,6 +747,7 @@ class RawTextRfcWriter(BaseRfcWriter):
         refkeys = []
         refsource = {}
         # [surname, initial.,] "title", (STD), (BCP), (RFC), (Month) Year.
+        i = 0
         for i, ref in enumerate(list.findall('reference')):
             refstring = []
             authors = ref.findall('front/author')
