@@ -7,10 +7,16 @@ import re
 import lxml.etree
 import datetime
 import traceback as tb
+
 from io import open
+from lxml import etree
+from lxml.etree import Element, Comment, CDATA
+
 from xml2rfc import log
-from xml2rfc.writers.base import default_options
-from lxml.etree import Element, Comment, CDATA, _Comment
+from xml2rfc.util.unicode import unicode_content_tags, isascii
+from xml2rfc.utils import find_duplicate_ids, hastext, isempty
+from xml2rfc.writers.base import default_options, BaseV3Writer
+
 
 try:
     import debug
@@ -27,7 +33,7 @@ def slugify(s):
 
 def idref(s):
     s = re.sub(r'^[^A-Za-z_]', '_', s)
-    s = re.sub(r'[^-._A-Za-z0-9]', '-', s)
+    s = re.sub(r'[^-._A-Za-z0-9]', '_', s)
     return s
 
 def stripattr(e, attrs):
@@ -40,30 +46,19 @@ def copyattr(a, b):
         v = a.get(k)
         b.set(k, v)
 
-def isempty(e):
-    return not ((len(e) and any([ not isinstance(c, _Comment) for c in e.iterchildren() ]))
-                or (e.text and e.text.strip()) or (e.tail and e.tail.strip()))
-
-def isblock(e):
-    return e.tag in [ 'artwork', 'dl', 'figure', 'ol', 'sourcecode', 't', 'ul', ]
-
-def iscomment(e):
-    return isinstance(e, _Comment)
-
-def hastext(e):
-    head = [ e.text ] if e.text and e.text.strip() else []
-    items = head + [ c for c in e.iterchildren() if not (isblock(c) or iscomment(c))] + [ c.tail for c in e.iterchildren() if c.tail and c.tail.strip() ]
-    return items
-
-class V2v3XmlWriter:
+class V2v3XmlWriter(BaseV3Writer):
     """ Writes an XML file with v2 constructs converted to v3"""
 
     def __init__(self, xmlrfc, quiet=None, options=default_options, date=datetime.date.today()):
+        super(V2v3XmlWriter, self).__init__(xmlrfc, quiet=quiet, options=options, date=date)
         if not quiet is None:
             options.quiet = quiet
         self.xmlrfc = xmlrfc
-        self.root = xmlrfc.getroot()
+        self.tree = xmlrfc.tree
+        self.root = self.tree.getroot()
         self.options = options
+        self.v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
+        self.schema = etree.ElementTree(file=self.v3_rng_file)
 
     def validate(self):
         v3_rng_file = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'data', 'v3.rng')
@@ -72,10 +67,10 @@ class V2v3XmlWriter:
             v3_rng.assertValid(self.root)
             log.note("The document validates according to the RFC7991 schema")
         except Exception as e:
-            # These warnings are occasionally incorrect -- disable this
-            # output for now:
-            #log.warn('\nInvalid document: %s' % (e,))
-            e = e
+            dups = find_duplicate_ids(self.schema, self.tree)
+            for attr, id, e in dups:
+                self.warn(e, 'Duplicate xsd:ID attribute %s="%s" found.  This will cause validation failure.' % (attr, id, ))
+            log.warn('\nInvalid document: %s' % (e,))
 
     def add_xinclude(self):
         for e in self.root.xpath('.//back//reference'):
@@ -86,24 +81,26 @@ class V2v3XmlWriter:
                     if num and num.isdigit():
                         xi = self.element('{http://www.w3.org/2001/XInclude}include',
                                     nsmap=self.xmlrfc.nsmap,
+                                    line=e.sourceline,
                                     href="https://xml2rfc.tools.ietf.org/public/rfc/bibxml/reference.RFC.%s.xml"%num)
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
                     else:
-                        self.warn('Invalid value in %s' % lxml.etree.tostring(e))
+                        self.warn(e, 'Invalid value in %s' % lxml.etree.tostring(e))
                 elif si.get('name') == 'Internet-Draft':
                     name = si.get('value', '')
                     if name:
                         tag = name[len('draft-'):] if name.startswith('draft-') else name
                         xi = self.element('{http://www.w3.org/2001/XInclude}include',
                                     nsmap=self.xmlrfc.nsmap,
+                                    line=e.sourceline,
                                     href="https://xml2rfc.tools.ietf.org/publicrfc/bibxml3/reference.I-D.%s.xml"%tag)
                         i = p.index(e)
                         p.remove(e)
                         p.insert(i, xi)
                     else:
-                        self.warn('Invalid value in %s' % lxml.etree.tostring(e))
+                        self.warn(e, 'Invalid value in %s' % lxml.etree.tostring(e))
         lxml.etree.cleanup_namespaces(self.root, top_nsmap=self.xmlrfc.nsmap)
 
 
@@ -132,19 +129,18 @@ class V2v3XmlWriter:
 
     # --- Element Operations -------------------------------------------
 
-    def element(self, tag, *args, **kwargs):
+    def element(self, tag, line=None, **kwargs):
         e = Element(tag, **kwargs)
-        if self.options.debug:
+        if line:
+            e.sourceline = line
+        elif self.options.debug:
             filename, lineno, caller, code = tb.extract_stack()[-2]
-            e.set('line', str(lineno))
-        if args:
-            assert len(args) == 1
-            text = args[0]
-            e.text = text
+            e.base = os.path.basename(filename)
+            e.sourceline = lineno
         return e
 
     def copy(self, e, tag):
-        n = self.element(tag)
+        n = self.element(tag, line=e.sourceline)
         n.text = e.text
         n.tail = e.tail
         copyattr(e, n)
@@ -163,29 +159,33 @@ class V2v3XmlWriter:
         if not isinstance(comments, list):
             comments = [comments]
         p = a.getparent()
-        i = p.index(a)
-        c = None
-        if self.options.verbose:
-            for comment in comments:
-                c = Comment(" v2v3: %s " % comment.strip())
-                p.insert(i, c)
-                i += 1
-        if not b is None:
-            if a.text and a.text.strip():
-                b.text = a.text
-            if a.tail and a.tail.strip():
-                b.tail = a.tail
-            copyattr(a, b)
-            for child in a.iterchildren():
-                b.append(child)         # moves child from a to b
-            p.replace(a, b)
-        else:
-            if a.tail:
-                if c is None:
-                    p.text = p.text + ' ' + a.tail if p.text else a.tail
-                else:
-                    c.tail = a.tail
-            p.remove(a)
+        if p != None:
+            i = p.index(a)
+            c = None
+            if self.options.verbose:
+                for comment in comments:
+                    c = Comment(" v2v3: %s " % comment.strip())
+                    p.insert(i, c)
+                    i += 1
+            if not b is None:
+                if a.text and a.text.strip():
+                    b.text = a.text
+                if a.tail != None:
+                    b.tail = a.tail
+                b.sourceline = a.sourceline
+                copyattr(a, b)
+                for child in a.iterchildren():
+                    b.append(child)         # moves child from a to b
+                p.replace(a, b)
+            else:
+                if a.tail:
+                    if c is None:
+                        p.text = p.text + ' ' + a.tail if p.text else a.tail
+                    else:
+                        c.tail = a.tail
+                p.remove(a)
+        if b != None:
+            b.sourceline = a.sourceline
         return b
 
     def move_after(self, a, b, comments=None):
@@ -220,7 +220,7 @@ class V2v3XmlWriter:
         assert t.tag == 't'
         pp = t.getparent()
         i = pp.index(t)+1
-        t2 = self.element('t')
+        t2 = self.element('t', line=e.sourceline)
         t2.text = e.tail
         e.tail = None
         for s in e.itersiblings():
@@ -235,8 +235,9 @@ class V2v3XmlWriter:
 
     def wrap_content(self, e, t=None):
         if t is None:
-            t = self.element('t')
+            t = self.element('t', line=e.sourceline)
         t.text = e.text
+        t.sourceline = e.sourceline
         e.text = None
         for c in e.iterchildren():
             t.append(c)
@@ -246,6 +247,9 @@ class V2v3XmlWriter:
     # ------------------------------------------------------------------
 
     def convert2to3(self):
+        if self.root.get('version') in ['3', ]:
+            return self.tree
+
         selectors = [
             # we need to process list before block elements that might get
             # promoted out of their surrounding <t/>, as <list/> uses one <t/>
@@ -257,6 +261,12 @@ class V2v3XmlWriter:
                                             # 2.5.4.  "height" Attribute
                                             # 2.5.8.  "width" Attribute
                                             # 2.5.9.  "xml:space" Attribute
+            './/back',
+            # We need to process preamble and postamble before figure,
+            # because artwork or sourcecode within a figure could later be
+            # promoted and the figure discarded.
+            './/postamble',                 # 3.5.  <postamble>
+            './/preamble',                  # 3.6.  <preamble>
             './/figure',                    # 2.25.  <figure>
                                             # 2.25.1.  "align" Attribute
                                             # 2.25.2.  "alt" Attribute
@@ -279,8 +289,6 @@ class V2v3XmlWriter:
                                             # 2.66.2.  "pageno" Attribute
             './/facsimile',                 # 3.2.  <facsimile>
             './/format',                    # 3.3.  <format>
-            './/postamble',                 # 3.5.  <postamble>
-            './/preamble',                  # 3.6.  <preamble>
             './/spanx',                     # 3.7.  <spanx>
             './/texttable',                 # 3.8.  <texttable>
             './/vspace',                    # 3.10.  <vspace>
@@ -297,10 +305,11 @@ class V2v3XmlWriter:
             './/*[self::artwork or self::dl or self::figure or self::ol or self::sourcecode or self::t or self::ul]',
             './/*[@*="yes" or @*="no"]',      # convert old attribute false/true
             '.;pretty_print_prep()',
+            '.;wrap_non_ascii()',
         ]
 
         # replace the vocabulary v2 dtd, but keep some entity definitions.
-        tree = self.root.getroottree()
+        tree = self.tree
         tree.docinfo.system_url = "rfc2629-xhtml.ent"
 
         for s in selectors:
@@ -324,6 +333,10 @@ class V2v3XmlWriter:
             else:
                 log.warn("No handler for function %s, slug %s" % (func_name, slug, ))
 
+        self.root.set('version', '3')
+
+        return self.tree
+
     # ----------------------------------------------------------------------
 
     # 1.3.2.  New Attributes for Existing Elements
@@ -332,6 +345,8 @@ class V2v3XmlWriter:
     #       to <rfc> to cover Processing Instructions (PIs) that were in v2
     #       that are still needed in the grammar.  ...
     def processing_instruction(self, e, p):
+        if e.target != 'rfc':
+            return
         rfc_element = self.root.find('.')
         pi_name = {
             'sortrefs': 'sortRefs',
@@ -346,6 +361,8 @@ class V2v3XmlWriter:
                     v = 'true'
                 rfc_element.set(a, v)
                 self.replace(e, None, 'Moved %s PI to <rfc %s="%s"' % (k, a, v))
+        else:
+            self.replace(e, None, 'Removed %s PI"' % (k,))
 
     # 1.3.4.  Additional Changes from v2
     # 
@@ -357,7 +374,8 @@ class V2v3XmlWriter:
             pass
         elif p.tag == 'reference':
             title = p.find('./front/title')
-            self.move_after(e, title, "Moved <seriesInfo/> inside <front/> element")
+            if title != None:
+                self.move_after(e, title, "Moved <seriesInfo/> inside <front/> element")
 
     # 1.3.4.  Additional Changes from v2
     #
@@ -367,8 +385,10 @@ class V2v3XmlWriter:
         # check if we have any text or non-block-elements next to the
         # element.  If so, embed in <t/> and then promote the element.
         nixmix_tags = [ 'blockquote', 'li', 'dd', 'td', 'th']
-        if p.tag in nixmix_tags and hastext(p):
-            #debug.say('Wrap %s content ...'%(p.tag))
+        text_items = hastext(p)
+        if p.tag in nixmix_tags and text_items:
+            if self.options.verbose:
+                self.warn(e, "Found mixed content in <%s>: Both <%s> and inline content: '%s'" % (p.tag, e.tag, (' '.join(str(t) for t in text_items))[:40].strip().replace('\n',' ')))
             self.wrap_content(p)
         p = e.getparent()
         #pp = p.getparent()
@@ -390,12 +410,14 @@ class V2v3XmlWriter:
     # 
     #    Deprecated.
     def element_artwork(self, e, p):
-        e.text = CDATA(e.text)          # prevent text from being mucked up by other processors
+        if e.text and e.text.strip():
+            e.text = CDATA(e.text)          # prevent text from being mucked up by other processors
         stripattr(e, ['height', '{http://www.w3.org/XML/1998/namespace}space', 'width', ])
-        if re.search(r'^\s*<CODE BEGINS>', e.text):
+        if e.text and re.search(r'^\s*<CODE BEGINS>', e.text):
             # We have source code.  Permitted attributes: anchor, name,
             # source, type.
             e = self.replace(e, 'sourcecode')
+            e.set('markers', 'true')
             match = re.search(r'(?s)^\s*<CODE BEGINS>(\s+file\s+"([^"]*)")?(.*?)(<CODE ENDS>(.*))?$', e.text)
             file = match.group(1)
             name = match.group(2)
@@ -406,12 +428,27 @@ class V2v3XmlWriter:
                 e.set('name', name)
             e.text = CDATA(body)
             if not ends:
-                self.warn("Found <CODE BEGINS> without matching <CODE ENDS>")
+                self.warn(e, "Found <CODE BEGINS> without matching <CODE ENDS>")
             if ends and tail.strip() != "":
-                self.warn("Found non-whitespace content after <CODE ENDS>")
+                self.warn(e, "Found non-whitespace content after <CODE ENDS>")
                 e.tail = tail
             stripattr(e, ['align', 'alt', 'height', 'suppress-title', 'width', ])
 
+    def element_back(self, e, p):
+        # XXXX The RFCs don't say anything about the text rendering of
+        # references.  The following transforms multiple <references> in
+        # <back> to be encapsulated within one <references>.
+        references = list(e.iterchildren('references'))
+        if len(references) > 1:
+            pos = e.index(references[0])
+            refs = self.element('references', line=references[0].sourceline)
+            name = self.element('name', line=references[0].sourceline)
+            name.text = "References"
+            refs.append(name)
+            e.insert(pos, refs)
+            for r in references:
+                refs.append(r)          # moves r
+
 
     # 2.25.  <figure>
     # 
@@ -452,18 +489,35 @@ class V2v3XmlWriter:
     # 
     def element_figure(self, e, p):
         comments = []
-        artwork = e.find('./artwork')
-        if not artwork is None:
-            for attr in ['alt', 'src', ]:
+        embedded = e.find('./artwork')
+        if embedded == None:
+            embedded = e.find('./sourcecode')
+        if embedded != None:
+            for attr in ['align', 'alt', 'src', ]:
                 if attr in e.attrib:
                     fattr = e.get(attr)
-                    if attr in artwork.attrib:
-                        aattr = artwork.get(attr)
+                    if attr in embedded.attrib:
+                        aattr = embedded.get(attr)
                         if fattr != aattr:
                             comments.append('Warning: The "%s" attribute on artwork differs from the one on figure.  Using only "%s" on artwork.' % (attr, attr))
                     else:
-                        artwork.set(attr, fattr)
-        stripattr(e, ['align', 'alt', 'height', 'src', 'suppress-title', 'width', ])
+                        if embedded.tag == 'artwork' or attr == 'src':
+                            embedded.set(attr, fattr)
+                    stripattr(e, [ attr ])
+            # if we have <artwork> and either no anchor and no title, or suppress-title='true',
+            # then promote the <artwork> and get rid of the <figure>
+        if embedded != None and ((not e.get('anchor') and (not e.get('title')) or e.get('suppress-title') == 'true')
+            and e.find('name')==None and e.find('preamble')==None and e.find('postamble')==None ):
+            pos = p.index(e)
+            embedded.tail = e.tail
+            p.remove(e)
+            p.insert(pos, embedded)
+        else:
+            stripattr(e, ['align', 'height', 'src', 'suppress-title', 'width', ])
+            if self.options.strict:
+                stripattr(e, ['alt', ])
+
+
 
     # 2.25.7.  "title" Attribute
     # 
@@ -485,7 +539,7 @@ class V2v3XmlWriter:
     # 
     #    Deprecated.
     def attribute_title(self, e, p):
-        n = self.element('name')
+        n = self.element('name', line=e.sourceline)
         n.text = e.get('title').strip()
         if n.text:
             e.insert(0, n)
@@ -532,23 +586,38 @@ class V2v3XmlWriter:
 #             "historic": "Historic",
 #             "info": "Informational",
 #         }
+        def equal(e1, e2):
+            return (e1.get('name'), e1.get('value')) ==  (e2.get('name'), e2.get('value'))
         front = e.find('./front')
         title = e.find('./front/title')
-        i = front.index(title) + 1
+        i = front.index(title) + 1 if title!=None else 0
+        series = front.xpath('seriesInfo')
         if 'category' in e.attrib and 'seriesNo' in e.attrib:
             attr = {
                 'name': e.get('category'),
                 'value': e.get('seriesNo')
             }
-            front.insert(i, self.element('seriesInfo', **attr))
+            new = self.element('seriesInfo', line=e.sourceline, **attr)
+            if not [ s for s in series if equal(s, new) ]:
+                front.insert(i, new)
             stripattr(e, ['seriesNo', ]) # keep 'category' for use in preptool
         if 'number' in e.attrib:
-            front.insert(i, self.element('seriesInfo', name="RFC", value=e.get('number')))
+            new = self.element('seriesInfo', name="RFC", value=e.get('number'), line=e.sourceline)
+            if not [ s for s in series if equal(s, new) ]:
+                front.insert(i, new)
             if 'docName' in e.attrib:
-                e.insert(0, self.element('link', rel='convertedFrom', href="https://datatracker.ietf.org/doc/%s"%(e.get('docName'), )))
+                e.insert(0, self.element('link', rel='prev', href="https://datatracker.ietf.org/doc/%s"%(e.get('docName'), ), line=e.sourceline))
         elif 'docName' in e.attrib:
-            front.insert(i, self.element('seriesInfo', name="Internet-Draft", value=e.get('docName')))
-        stripattr(e, ['docName', 'number', 'xi'])
+            value=e.get('docName')
+            if '.' in value:
+                log.warn("The 'docName' attribute of the <rfc/> element should not contain any filename extension: docName=\"draft-foo-bar-02\".")
+            if not re.search('-\d\d$', value):
+                log.warn("The 'docName' attribute of the <rfc/> element should have a revision number as the last component when submitted: docName=\"draft-foo-bar-02\".")
+            new = self.element('seriesInfo', name="Internet-Draft", value=value, line=e.sourceline)
+            if not 'Internet-Draft' in [ s.get('name') for s in series ]:
+                front.insert(i, new)
+
+        stripattr(e, ['xi', ])
 
     # 2.47.  <seriesInfo>
     # 
@@ -566,7 +635,7 @@ class V2v3XmlWriter:
     def element_t(self, e, p):
         if p.tag != 'list':
             stripattr(e, ['hangText', ])
-            
+
     # 
     # 
     # 2.66.  <xref>
@@ -629,8 +698,11 @@ class V2v3XmlWriter:
         if ptarget:
             if ftarget == ptarget:
                 self.replace(e, None, "<format/> element with duplicate target (%s) removed" % ftarget)
-            else:
-                self.replace(e, None, "Warning: <format/> element with alternative target (%s) removed" % ftarget)
+            # The following, while implementing RFC 7991, seems dubious, and
+            # has been commented out.  Why would you disallow pointing to
+            # multiple versions of a reference?
+            #else:
+            #    self.replace(e, None, "Warning: <format/> element with alternative target (%s) removed" % ftarget)
         else:
             p.set('target', ftarget)
             self.replace(e, None, "<format/> element removed, target value (%s) moved to parent" % ftarget)
@@ -671,11 +743,13 @@ class V2v3XmlWriter:
                 elif parent.tag == 'ul':
                     style = 'symbols'
                 elif parent.tag == 'ol':
-                    nstyle = parent.get('style')
-                    if nstyle in ['a', 'A']:
-                        style = 'letters'
-                    elif nstyle == '1':
-                        style = 'numbers'
+                    style = 'inherit'
+                    nstyle = parent.get('type')
+                    # alternate letter case in sub-lists:
+                    if   nstyle in 'ai':
+                        nstyle = nstyle.upper()
+                    elif nstyle in 'AI':
+                        nstyle = nstyle.lower()                        
                 if style:
                     break
         if not style:
@@ -683,11 +757,13 @@ class V2v3XmlWriter:
         #
         if   style == 'empty':
             tag = 'ul'
+            attribs['empty'] = 'true'
         elif style == 'symbols':
             tag = 'ul'
         elif style == 'hanging':
             tag = 'dl'
-        elif style == 'numbers':
+            attribs["newline"] = "false"
+        elif style in ['numbers', 'inherit']:
             tag = 'ol'
             attribs['type'] = nstyle if nstyle else '1'
         elif style == 'letters':
@@ -701,30 +777,40 @@ class V2v3XmlWriter:
             comments.append("Warning: unknown list style: '%s'" % style)
         #
         comments.append('Replaced <list style="%s"/> with <%s/>' % (style, tag))
-        if 'counter' in e.keys():
+        if tag=='ol' and 'counter' in e.keys():
             attribs['group'] = e.get('counter')
             comments.append("Converting <list counter=...> to <%s group=...> " % tag)
         #
-        attribs['spacing'] = 'compact' if hasattr(e, 'pis') and e.pis['compact'] in ['yes', 'true'] else 'normal'
+        attribs['spacing'] = 'compact' if hasattr(e, 'pis') and e.pis['subcompact'] in ['yes', 'true'] else 'normal'
         #
         stripattr(e, ['counter', 'style', ])
         l = self.element(tag, **attribs)
         self.replace(e, l, comments)
+        indent = l.get('hangIndent')
         stripattr(l, ['hangIndent'])
         if tag in ['ol', 'ul']:
             for t in l.findall('./t'):
                 self.replace(t, 'li')
         elif tag == 'dl':
+            if indent:
+                l.set('indent', indent)
             for t in l.findall('./t'):
-                dt = self.element('dt')
+                dt = self.element('dt', line=t.sourceline)
                 dt.text = t.get('hangText')
                 if not dt.text is None:
                     del t.attrib['hangText']
+                # Convert <vspace> at the start of hanging list text to
+                # attribute hanging='true' on <dl>
+                if not t.text or not t.text.strip():
+                    if len(t) and t[0].tag == 'vspace':
+                        t.text = t[0].tail
+                        t.remove(t[0])
+                        l.set('newline', 'true')
                 i = l.index(t)
                 l.insert(i, dt)
                 self.replace(t, 'dd')
         else:
-            self.error("Unexpected tag when processing <list/>: '%s'" % tag)
+            self.warn(e, "Unexpected tag when processing <list/>: '%s'" % tag)
 
     # 3.4.1.  "counter" Attribute
     # 
@@ -806,7 +892,8 @@ class V2v3XmlWriter:
         if style in tags:
             tag = tags[style]
         else:
-            self.error("Unexpected style in <spanx/>: '%s'" % style)
+            self.warn(e, "Unexpected style in <spanx/>: '%s'" % style)
+            tag = None
         stripattr(e, ['style', '{http://www.w3.org/XML/1998/namespace}space', ])
         self.replace(e, tag, 'Replaced <spanx style="%s"/> with <%s/>' % (style, tag))
 
@@ -847,7 +934,7 @@ class V2v3XmlWriter:
         tbody = None
         tr = None
         align = []
-        table = self.element('table')
+        table = self.element('table', line=e.sourceline)
         copyattr(e, table)
         for x in e.iterchildren():
             if   x.tag == 'preamble':
@@ -855,9 +942,9 @@ class V2v3XmlWriter:
                 pass
             elif x.tag == 'ttcol':
                 if colcount == 0:
-                    thead = self.element('thead')
+                    thead = self.element('thead', line=x.sourceline)
                     table.append(thead)
-                    tr = self.element('tr')
+                    tr = self.element('tr', line=x.sourceline)
                     thead.append(tr)
                 align.append(x.get('align', None))
                 th = self.copy(x, 'th')
@@ -867,10 +954,10 @@ class V2v3XmlWriter:
             elif x.tag == 'c':
                 col = cellcount % colcount
                 if cellcount == 0:
-                    tbody = self.element('tbody')
+                    tbody = self.element('tbody', line=x.sourceline)
                     table.append(tbody)
                 if col == 0:
-                    tr = self.element('tr')
+                    tr = self.element('tr', line=x.sourceline)
                     tbody.append(tr)
                 td = self.copy(x, 'td')
                 if align[col]:
@@ -881,7 +968,8 @@ class V2v3XmlWriter:
             elif x.tag == 'postamble':
                 # will be handled separately
                 pass
-        stripattr(table, ['align', 'style', 'suppress-title',])
+        stripattr(table, ['style', 'suppress-title',])
+        #stripattr(table, ['style', 'suppress-title',])
         p.replace(e, table)
 
 
@@ -932,7 +1020,7 @@ class V2v3XmlWriter:
         l = t.getparent()
         if l.tag in ['dd', 'li', ]:
             i = l.index(t) + 1
-            t2 = self.element('t')
+            t2 = self.element('t', line=e.sourceline)
             if e.tail and e.tail.strip():
                 t2.text = e.tail
             for s in e.itersiblings():
@@ -983,3 +1071,67 @@ class V2v3XmlWriter:
                 e.set(k, 'true')
             elif v == 'no':
                 e.set(k, 'false')
+
+#     def pretty_print_prep(self, e, p):
+#         # apply this to elements that can't appear with text, i.e., don't have
+#         # any of these as parent:
+#         skip_parents = set([
+#             "annotation", "blockquote", "preamble", "postamble", "name", "refcontent", "c", "t",
+#             "cref", "dd", "dt", "li", "td", "th", "tt", "em", "strong", "sub", "sup", ])
+#         for c in e.iter():
+#             p = c.getparent()
+#             if p != None and p.tag in skip_parents:
+#                 continue
+#             if c.tail != None:
+#                 if c.tail.strip() == '':
+#                     c.tail = None
+
+    def wrap_non_ascii(self, e, p):
+        self.downcode_punctuation()
+        for e in self.tree.iter():
+            def uwrap(text, line): 
+                words = []
+                elements = []
+                ascii = None
+                for word in re.split('(\s+)', text, flags=re.U):
+                    if isascii(word):
+                        words.append(word)
+                    else:
+                        u = self.element('u', line=line)
+                        u.text = word
+                        if ascii is None:
+                            ascii = ''.join(words)
+                            words = []
+                            elements.append(u)
+                        else:
+                            elements[-1].tail = ''.join(words)
+                            words = []
+                            elements.append(u)
+                if words:
+                    if ascii is None:
+                        ascii = ''.join(words)
+                    else:
+                        elements[-1].tail = ''.join(words)
+                return ascii, elements
+            #
+            if e.text and e.tag not in unicode_content_tags:
+                    try:
+                        e.text.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.text, elements = uwrap(e.text, e.sourceline)
+                        if len(e):
+                            c = e[0]
+                            for u in elements:
+                                c.addprevious(u)
+                        else:
+                            for u in elements:
+                                e.append(u)
+            if e.tail and e.getparent().tag not in unicode_content_tags:
+                    try:
+                        e.tail.encode('ascii')
+                    except UnicodeEncodeError:
+                        e.tail, elements = uwrap(e.tail, e.sourceline)
+                        for u in elements:
+                            e.addnext(u)
+                            e = u
+                                
